/*______________________________________________________________________________
 * 
 * org.eidola.kernel.Capsule
 * 
 * Part of the Eidola Kernel Reference Implementation
 * See http://eidola.org for oodles of relevant fun!
 * 
 *______________________________________________________________________________
 * 
 * Copyright 1998, 2000-2001 Paul Cantrell
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License version 2, as published by the
 * Free Software Foundation. See the file LICENSE.html for more information.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY, including the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc. / 59 Temple
 * Place, Suite 330 / Boston, MA 02111-1307 / USA.
 *_______________________________________________________________________________
 */

package org.eidola.kernel;

import org.eidola.kernel.event.*;
import org.eidola.kernel.event.EventListener;
import org.eidola.kernel.error.*;
import org.eidola.util.CollectionDiff;

import java.util.*;

/**
	An element which encapsulates a unit of a program by holding
	member elements in public and private sections.
	<p>
	<font size=+1><b>Structure:</b></font>
	<ul>
		<li> name
		<li> owner
		<li> public members
		<li> private members
	</ul>
	<p>
	<font size=+1><b>Signature:</b></font>
	<ul>
		<li> name
		<li> public members
		<li> signatures of public members
	</ul>
	<p>
	See note on synchronization and concurrent read safety in {@link Container}.
	
	@author Paul Cantrell
	@version [Development version]
*/

public abstract class Capsule extends NamedElement
	{
	//------------------------------------------------------
	//  Constants
	//------------------------------------------------------
	
	/**	Part of a capsule.
	 *	@see org.eidola.kernel.event.StructureChanged */
	
	static public final ContainerPart
		 PUBLIC_MEMBERS = new ContainerPart( "public members"),
		PRIVATE_MEMBERS = new ContainerPart("private members");
	
	//------------------------------------------------------
	//  Constructors
	//------------------------------------------------------
	
	/** Creates a new empty capsule. */
	
	public Capsule()
		{
		super();
		pub = new HashSet();
		priv = new HashSet();
		memberListener = new EventListener()
			{
			public void handleEvent(Event event)
				{ handleContentEvent((ContainerEvent) event); }
			};
		}
	
	//------------------------------------------------------
	//  Structural properties
	//------------------------------------------------------
	
	/** Returns the public members of this capsule. */
	
	public Set/*<NamedElement>*/ getPublic()
		{ return pub; }
	
	/** Returns the private members of this capsule. */
	
	public Set/*<NamedElement>*/ getPrivate()
		{ return priv; }
	
	/** Finds a public member by name.  If there are illegal duplicate names,
	 *  this method returns one of them in an undefined fashion.
	 *  @param memName The member name to look for. */
	
	public NamedElement findPublic(String memName)
		{ return findMember(memName, true); }
	
	/** Finds a private member by name.  If there are illegal duplicate names,
	 *  this method returns one of them in an undefined fashion.
	 *  @param memName The member name to look for. */
	
	public NamedElement findPrivate(String memName)
		{ return findMember(memName, false); }
	
	/** Finds a member by name.  If there are illegal duplicate names,
	 *  this method returns one of them in an undefined fashion.
	 *  @param memName The member name to look for.
	 *  @param inPublic Search in public (true) or private (false) members. */
	
	public NamedElement findMember(String memName, boolean inPublic)
		{
		for(Iterator i = (inPublic ? getPublic() : getPrivate()).iterator(); i.hasNext(); )
			{
			NamedElement cur = (NamedElement) i.next();
			if(cur.getName().equals(memName))
				return cur; // found
			}
		return null; // not found
		}
	
	/** Adds a public member to this capsule.  If the element is already a
	 *  public member, this method has no effect.
	 *  <p><b>Events:</b> {@link StructureChanged}, {@link SignatureChanged}
	 *  @param mem The member to add.
	 *  @throws IllegalElementType If mem is null. */
	
	public void addPublic(NamedElement mem)
		throws IllegalElementType
		{ addMember(mem, true); }
		
	/** Adds a private member to this capsule.  If the element is already a
	 *  private member, this has no effect.
	 *  <p><b>Events:</b> {@link StructureChanged}, {@link SignatureChanged}
	 *  @param mem The member to add.
	 *  @throws IllegalElementType If mem is null. */
	
	public void addPrivate(NamedElement mem)
		throws IllegalElementType
		{ addMember(mem, false); }
		
	/** Adds a member to this capsule.  If it is already a member where you are adding
	 *  it, this method has no effect.
	 *  <p><b>Events:</b> {@link StructureChanged}, {@link SignatureChanged}
	 *  @param mem The member to add.
	 *  @param inPublic Makes the element a public (true) or private (false) member.
	 *  @throws IllegalElementType If mem is null. */
	
	public void addMember(NamedElement mem, boolean inPublic)
		throws IllegalElementType
		{
		checkMemberType(mem);
		if(mem == null)
			throw new NullPointerException("mem == null");
		if((inPublic ? pub : priv).contains(mem))
			return;
		
		Set newMems = new HashSet(inPublic ? pub : priv);
		newMems.add(mem);
		newMems = Collections.unmodifiableSet(newMems);
		if(inPublic)
			pub = newMems;
		else
			priv = newMems;
		
		mem.addListener(memberListener, Engine.getInstance().getPropagatorQ());
		StructureChanged structEvent =
			new StructureChanged(
				this,
				inPublic ? PUBLIC_MEMBERS : PRIVATE_MEMBERS);
		broadcastEvent(structEvent);
		if(inPublic)
			broadcastEvent(new SignatureChanged(this, structEvent));
		}
	
	/** Removes a public member from this capsule.  If the element is not a public
	 *  member, this method has no effect.
	 *  <p><b>Events:</b> {@link StructureChanged}, {@link SignatureChanged}
	 *  @param mem The member to remove. */
	
	public void removePublic(NamedElement mem)
		{ removeMember(mem, true); }
		
	/** Removes a private member from this capsule.  If the element is not a private
	 *  member, this method has no effect.
	 *  <p><b>Events:</b> {@link StructureChanged}, {@link SignatureChanged}
	 *  @param mem The member to remove. */
	
	public void removePrivate(NamedElement mem)
		{ removeMember(mem, false); }
		
	/** Removes a member from this capsule.  If the element is not a
	 *  member, this method has no effect.
	 *  <p><b>Events:</b> {@link StructureChanged}, {@link SignatureChanged}
	 *  @param mem The member to remove. 
	 *  @param inPublic Removes the element from the public (true) or private (false) members. */
	
	public void removeMember(NamedElement mem, boolean inPublic)
		{
		if(mem == null)
			throw new NullPointerException("mem == null");
		
		Set newMems = new HashSet(inPublic ? pub : priv);
		newMems.remove(mem);
		newMems = Collections.unmodifiableSet(newMems);
		if(inPublic)
			pub = newMems;
		else
			priv = newMems;
			
		if(!(inPublic ? priv : pub).contains(mem)) // Mem still might be in other half
			mem.removeListener(memberListener);
		StructureChanged structEvent =
			new StructureChanged(
				this,
				inPublic ? PUBLIC_MEMBERS : PRIVATE_MEMBERS);
		broadcastEvent(structEvent);
		if(inPublic)
			broadcastEvent(new SignatureChanged(this, structEvent));
		}
	
	/** Sets the public members of this capsule.
	 *  <p><b>Events:</b> {@link StructureChanged}, {@link SignatureChanged}
	 *  @param newPub A collection of {@link NamedElement}s which will be the new
	 *  	public members. This method makes an internal copy of this collection,
	 *      so it is safe to pass a collection and subsequently modify it.
	 *  @throws IllegalElementType If one of the new members is null or not a named element. */
	
	public void setPublic(Collection/*<NamedElement>*/ newPub)
		throws IllegalElementType
		{ setMembers(newPub, true); }
		
	/** Sets the private members of this capsule.
	 *  <p><b>Events:</b> {@link StructureChanged}, {@link SignatureChanged}
	 *  @param newPriv A collection of {@link NamedElement}s which will be the new
	 *  	private members. This method makes an internal copy of this collection,
	 *      so it is safe to pass a collection and subsequently modify it.
	 *  @throws IllegalElementType If one of the new members is null or not a named element. */
	
	public void setPrivate(Collection/*<NamedElement>*/ newPriv)
		throws IllegalElementType
		{ setMembers(newPriv, false); }
	
	/** Sets either the public or private members of this capsule.
	 *  <p><b>Events:</b> {@link StructureChanged}, {@link SignatureChanged}
	 *  @param newMems A collection of {@link NamedElement}s which will be the new
	 *  	members. This method makes an internal copy of this collection,
	 *      so it is safe to pass a collection and subsequently modify it.
	 *  @param setPub Set the public (true) or private (false) members.
	 *  @throws IllegalElementType If one of the new members is null or not a named element. */
	
	public void setMembers(Collection/*<NamedElement>*/ newMems, boolean setPub)
		throws IllegalElementType
		{
		if(newMems == null)
			newMems = new HashSet();
		for(Iterator m = newMems.iterator(); m.hasNext(); )
			checkMemberType(m.next());
		Set mems = Collections.unmodifiableSet(new HashSet(newMems));
		
		EventBroadcaster.updateBroadcasters(
			memberListener,
			Engine.getInstance().getPropagatorQ(),
			setPub ? pub : priv,
			mems);

		if(setPub)
			pub = mems;
		else
			priv = mems;
		
		StructureChanged structEvent =
			new StructureChanged(
				this,
				setPub ? PUBLIC_MEMBERS : PRIVATE_MEMBERS);
		broadcastEvent(structEvent);
		if(setPub)
			broadcastEvent(new SignatureChanged(this, structEvent));
		}
	
	//------------------------------------------------------
	//  Semantic conditions and errors
	//------------------------------------------------------
	
	public abstract class Compilation
		extends NamedElement.Compilation
		{
		/** Calculates the contents of this capsule. */
		
		protected void calculateDerived()
			{
			contents = new HashSet();
			contents.addAll(pub);
			contents.addAll(priv);
			contents = Collections.unmodifiableSet(contents);
			
			elementsUsed = new HashSet(contents);
			elementsUsed.add(Capsule.this);
			elementsUsed.add(getOwner());
			elementsUsed = Collections.unmodifiableSet(elementsUsed);
			}
		
		/** Check the capsule structure.  A capsule must satisfy
		 *  the following conditions, in addition to the conditions required by
		 *  {@link Element.Compilation#checkLazy()}:
		 *  <ul>
		 *  	<li> No element can be both a public and a private member.
		 *  	<li> Public members' owners must generalize this capsule (this
		 *  		handles both class inheritance and package membership properly --
		 *  		see {@link Class.Compilation#specializes(NamedElement) Class.specializes()} and
		 *  		{@link Package.Compilation#specializes(NamedElement) Package.specializes()}).
		 *  	<li> If we're inheriting a member, it must be a public member of its owner.
		 *  	<li> Private members' owners must be this capsule.
		 *  </ul> */
		
		protected void checkLazy()
			{
			super.checkLazy();
			
			Set pub = getPublic();
			Set priv = getPrivate();
			
			// Public and private shouldn't overlap
			
			Set mpd = new HashSet(pub);
			mpd.retainAll(priv);
			if(!mpd.isEmpty())
				for(Iterator i = mpd.iterator(); i.hasNext(); )
					{
					Element curElem = (Element) i.next();
					addError(
						new PublicPrivateOverlap(
							curElem,
							curElem + " cannot be both a public and a private member of " + Capsule.this));
					}
			
			// Private members' owners must be this capsule
			
			for(Iterator i = getPrivate().iterator(); i.hasNext(); )
				{
				Element cur = (Element) i.next();
				Container curOwner = cur.getOwner();
				if(curOwner != Capsule.this)
					addError(new IllegalContent(
						cur,
						cur + " cannot be a private member of "
							+ Capsule.this + " because its owner is " + curOwner));
				}
			
			// Public members' owners must generalize this capsule
			
			for(Iterator i = getPublic().iterator(); i.hasNext(); )
				{
				Element cur = (Element) i.next();
				Container curOwner = cur.getOwner();
				if(curOwner == null
				  || !(curOwner instanceof Capsule)
				  || !specializes((Capsule) curOwner))
					addError(
						new IllegalInheritedMember(
							cur,
							cur + " cannot be a public member of " + Capsule.this
								+ " because its owner (" + curOwner
								+ ") does not generalize " + Capsule.this));
								
				// We can only inherit from generalization's public members
				
				else if(curOwner != Capsule.this && !((Capsule) curOwner).getPublic().contains(cur))
					addError(
						new IllegalInheritedMember(
							cur,
							Capsule.this + " cannot inherit non-public member " + cur));
				}
			}
		}
	
	//------------------------------------------------------
	//  Private
	//------------------------------------------------------
	
	private Set/*<NamedElement>*/ pub, priv;
	private EventListener memberListener;
	
	private void checkMemberType(Object member)
		throws IllegalElementType
		{
		if(member == null || !(member instanceof NamedElement))
			throw new IllegalElementType(
				(member instanceof Element) ? (Element) member : null,
				member + " cannot be a member of " + this
					+ " because it is not a package, class, function or variable");
		}
	}

	
