/*______________________________________________________________________________
 * 
 * org.eidola.kernel.Class
 * 
 * Part of the Eidola Kernel Reference Implementation
 * See http://eidola.org for oodles of relevant fun!
 * 
 *______________________________________________________________________________
 * 
 * Copyright 1998, 2000-2001 Paul Cantrell
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License version 2, as published by the
 * Free Software Foundation. See the file LICENSE.html for more information.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY, including the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc. / 59 Temple
 * Place, Suite 330 / Boston, MA 02111-1307 / USA.
 *_______________________________________________________________________________
 */

package org.eidola.kernel;

import org.eidola.kernel.event.*;
import org.eidola.kernel.event.EventListener;
import org.eidola.kernel.error.*;
import org.eidola.util.*;

import java.util.*;

/**
	An Eidola type.
	<p>
	Eidola determines type by the generalization relationship:
	a class explicitly declares which other classes are its <i>parents</i>,
	and the generalization relationship is then a transitive, reflexive relation over
	the chain of parents; see {@link Compilation#getGeneralizations()}.
	<p>
	For a class to legally claim another as its generalization, it must have a
	<i>matching subsignature.</i>  This means that (this is the Big Important Rule
	here) <b>for every public member of every parent class, there must be a matching
	public member of the child class</b>.
	<p>
	Familiar object-oriented languages will implicitly <i>inherit</i> members,
	but the Big Important Rule means that in Eidola, subclasses explicitly
	declare when they are inheriting a member.  To allow inheritance, the public
	members of a class fall under a slightly relaxed membership rule:  instead of
	requiring that a class own all of its public members, any <i>generalization</i>
	of the class may own them.  It is therefore possible for a class to borrow members
	from its generalizations.
	<p>
	Private members necessarily get inherited from generalizations, but this only
	happens at runtime and is invisible to static typing and scoping, so
	it's safe to ignore it for now.
	<p>
	A class can claim multiple parents.  In fact, since all potentially ambiguous
	inheritance happens explicitly, it is fine to have a cycle in the generalization
	graph.  Yes, it seems strange, but it's fine when you work through it -- all this
	means is that all the classes in the cycle have identical signatures.
	<p>
	<font size=+1><b>Structure:</b></font>
	<ul>
		<li> name
		<li> owner
		<li> public members
		<li> private members
		<li> parents
	</ul>
	<p>
	<font size=+1><b>Signature:</b></font>
	<ul>
		<li> name
		<li> signatures of public members
	</ul>
	<p>
	See note on synchronization and concurrent read safety in {@link Container}.
	
	@author Paul Cantrell
	@version [Development version]
*/

public class Class
	extends Capsule
	{
	//------------------------------------------------------
	//  Constants
	//------------------------------------------------------
	
	/**	Part of an element.
	 *	@see org.eidola.kernel.event.StructureChanged */
	
	static public final ContainerPart PARENTS = new ContainerPart("parents");
	
	//------------------------------------------------------
	//  Constructors
	//------------------------------------------------------
	
	/** Creates a new empty class. */
	
	public Class()
		{
		parents = new HashSet();
		parentListener = new EventListener()
			{
			public void handleEvent(Event event)
				{ handleParentEvent((ContainerEvent) event); }
			};
		}
		
	//------------------------------------------------------
	//  Structural properties
	//------------------------------------------------------
	
	/** Returns the set of parents, the generalizations this class explicitly declares.
	 *  @see Compilation#getGeneralizations() */
	
	public Set/*<Class>*/ getParents()
		{ return parents; }
	
	/** Changes the set of parents.
	 *  <p><b>Events:</b> {@link StructureChanged}
	 *  @param newParents A collection of {@link Class}es which will be the new
	 *  	parents. This method makes in internal copy of this collection,
	 *      so it is safe to pass a collection and subsequently modify it. */
	
	public void setParents(Collection/*<Class>*/ newParents)
		throws IllegalElementType
		{
		if(newParents == null)
			newParents = new HashSet();
		
		for(Iterator i = newParents.iterator(); i.hasNext(); )
			checkParentType(i.next());
		
		EventBroadcaster.updateBroadcasters(
			parentListener,
			Engine.getInstance().getPropagatorQ(),
			parents,
			newParents);
		
		parents = Collections.unmodifiableSet(new HashSet(newParents));
		
		broadcastEvent(new StructureChanged(this, PARENTS));
		}
	
	/** Adds a parent.
	 *  <p><b>Events:</b> {@link StructureChanged}
	 *  @param parent The class to add. */
	
	public void addParent(Class parent)
		throws IllegalElementType
		{
		checkParentType(parent);
		
		HashSet newParents = new HashSet(parents);
		newParents.add(parent);
		parents = Collections.unmodifiableSet(newParents);
		
		parent.addListener(parentListener, Engine.getInstance().getPropagatorQ());
		
		broadcastEvent(new StructureChanged(this, PARENTS));
		}
		
	/** Removes a parent.
	 *  <p><b>Events:</b> {@link StructureChanged}
	 *  @param parent The class to remove. */
	
	public void removeParent(Class parent)
		{
		if(parent == null)
			throw new NullPointerException("parent == null");
		
		HashSet newParents = new HashSet(parents);
		newParents.remove(parent);
		parents = Collections.unmodifiableSet(newParents);
		
		parent.removeListener(parentListener);
		
		broadcastEvent(new StructureChanged(this, PARENTS));
		}
	
	//------------------------------------------------------
	//  Changes and events
	//------------------------------------------------------
	
	/** Handles an event passed on from a member.  This handler
	 *  passes {@link SignatureChanged} events up the chain of public members,
	 *  and calls {@link Element#handleContentEvent(ContainerEvent)}. */
	
	public void handleContentEvent(ContainerEvent event)
		{
		super.handleContentEvent(event);
		
		if(event instanceof CompileCompleted)
			{
			CompileRequired prop = (CompileRequired) event.getPropagatedFrom();
			if(prop instanceof SignatureChanged)    // If the signature changed...
				{
				Container source = prop.getSource();
				if(getPublic().contains(source))     // ...for the public member...
					{
					ContainerEvent sigChanged = new SignatureChanged(this, prop);
					if(!prop.propagationChainContains(sigChanged))
						synchronized(this)
							{ broadcastEvent(sigChanged); }
					}
				}
			}
		}
	
	/** Handles an event passed on from a parent.  This handler
	 *  broadcasts a {@link CompileRequired} event when it receives a
	 *  {@link SignatureChanged} event from a parent. */
	
	public void handleParentEvent(ContainerEvent event)
		{
		// If a parent's signature has changed, the set of
		// elements we need to inherit/override may have changed
		
		if(event instanceof SignatureChanged)
			synchronized(this)
				{ broadcastEvent(new CompileRequired(this, (SignatureChanged) event)); }
		
		// If a parent's generalizations have changed, recompile
		// in case that's affected ours
		
		else if(event instanceof GeneralizationsChanged)
			synchronized(this)
				{ broadcastEvent(new CompileRequired(this)); } // Remember propogation?
		}
	
	//------------------------------------------------------
	//  Semantic conditions and errors
	//------------------------------------------------------
	
	protected Container.Compilation makeNewCompilation()
		{ return new Compilation(); }
		
	public class Compilation
		extends Capsule.Compilation
		{
		/** Creates a compilation. */
		
		public Compilation()
			{ generalizations = Collections.EMPTY_SET; }

		/** Returns true if the other class is a generalization of this class. */
		
		public boolean specializes(NamedElement other)
			{
			return other == Class.this
				|| ((other instanceof Class) && getGeneralizations().contains(other));
			}

		/** Returns the set which includes this class, its parent,
		 *  their parents, and so on.
		 *  @see Class#getParents() */
		
		public Set getGeneralizations()
			{ return generalizations; }
		
		/** Calculates the elements used and the generalizations.  If and only if
		 *  the set of generalizations changes, this method broadcasts a 
		 *  {@link GeneralizationsChanged} event. */
		
		protected void calculateDerived()
			{
			super.calculateDerived();
			
			// Add parents to elements used
			
			elementsUsed = new HashSet(elementsUsed);
			elementsUsed.addAll(parents);
			elementsUsed = Collections.unmodifiableSet(elementsUsed);
			
			// Recalculate generalizations
			
			Set newGeneralizations = Collections.unmodifiableSet(
				GraphUtil.reachableNodes(
					Collections.singleton(Class.this),
					new GraphWalker()
						{
						public Collection getEdgesFrom(Object node)
							{ return ((Class) node).getParents(); }
						} ));
			
			// If generalizations have changed, tell the world
			
			Set oldGeneralizations = ((Class.Compilation) getCompilation()).getGeneralizations();
			if(!newGeneralizations.equals(oldGeneralizations))
				synchronized(Class.this)
					{
					broadcastEvent(
						new GeneralizationsChanged( // Remember prop? How?
							Class.this,
							new CollectionDiff(oldGeneralizations, newGeneralizations)));
					}
			
			generalizations = newGeneralizations;
			}
		
		/** Checks the class structure.  A class must satisfy
		 *  the following condition, in addition to the conditions required by
		 *  {@link Capsule.Compilation#checkLazy()}:
		 *  <ul>
		 *  	<li> For every public member <i>A</i> of every parent, there must
		 *  		be a public member member <i>B</i> of this class such that
		 *          B {@link NamedElement.Compilation#specializes(NamedElement) specializes} A
		 *  		and has the same name.
		 *  </ul>
		 *  @return A list of {@link SemanticViolation}s, or an empty list if the structure is OK. */
		
		protected void checkLazy()
			{
			super.checkLazy();
			
			// For all public members of all parents... 
			
			for(Iterator parent = getParents().iterator(); parent.hasNext(); )
				for(Iterator pmem = ((Class) parent.next()).getPublic().iterator(); pmem.hasNext(); )
					{
					// ...there should be a matching public member in this class.
					
					boolean found = false;
					NamedElement curpmem = (NamedElement) pmem.next();
					
					for(Iterator memi = getPublic().iterator(); !found && memi.hasNext(); )
						{
						NamedElement mem = (NamedElement) memi.next();
						if(mem.getName().equals(curpmem.getName())
						  && ((NamedElement.Compilation) mem.getCompilation()).specializes(curpmem))
							found = true;
						}

					if(!found)
						addError(
							new MissingInheritedMember(
								curpmem, "No member in " + Class.this
								 + " to match inherited member " + curpmem));
					}
			}
		
		private Set/*<Class>*/ generalizations;
		}
	
	//------------------------------------------------------
	//  Debugging
	//------------------------------------------------------
	
	public void dump(int i)
		{
		Util.indent(i);
		
		System.out.print("Class " + toString());
		if(!getCompilation().isCurrent())
			System.out.print(" ");
		if(!parents.isEmpty())
			{
			System.out.print(" extends");
			for(Iterator elem = parents.iterator(); elem.hasNext(); )
				{
				System.out.print(" " + elem.next());
				if(elem.hasNext())
					System.out.print(",");
				}
			}
		System.out.println("");

		Util.indent(i+2);
		System.out.print("public:\n");
		for(Iterator elemi = getPublic().iterator(); elemi.hasNext(); )
			{
			Element elem = (Element) elemi.next();
			if(elem.getOwner() == this)
				elem.dump(i+4);
			else
				{
				Util.indent(i+4);
				System.out.println(elem + " (inherited)");
				}
			}

		Util.indent(i+2);
		System.out.print("private:\n");
		for(Iterator elem = getPrivate().iterator(); elem.hasNext(); )
			((Element) elem.next()).dump(i+4);
		}

	//------------------------------------------------------
	//  Private
	//------------------------------------------------------
	
	private Set/*<Class>*/ parents;
	private EventListener parentListener;
	
	private void checkParentType(Object parent)
		throws IllegalElementType
		{
		if(parent == null || !(parent instanceof Class))
			throw new IllegalElementType(
				(parent instanceof Element) ? (Element) parent : null,
				parent + " cannot be a parent of " + this
					+ " because it is not a class");
		}
	}

