/*______________________________________________________________________________
 * 
 * org.eidola.kernel.Function
 * 
 * Part of the Eidola Kernel Reference Implementation
 * See http://eidola.org for oodles of relevant fun!
 * 
 *______________________________________________________________________________
 * 
 * Copyright 1998, 2000-2001 Paul Cantrell
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License version 2, as published by the
 * Free Software Foundation. See the file LICENSE.html for more information.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY, including the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc. / 59 Temple
 * Place, Suite 330 / Boston, MA 02111-1307 / USA.
 *_______________________________________________________________________________
 */

package org.eidola.kernel;

import org.eidola.kernel.event.*;
import org.eidola.kernel.event.EventListener;
import org.eidola.kernel.error.*;
import org.eidola.util.Util;
import java.util.*;

/**
	Holds a body of code which takes input and output parameters.
	<p>
	A function's parameters are {@link Variable}s, which means that they are
	named, and they come in a list instead of a set, which means that they have an
	order.  This is to allow a UI to show parameter passing by order
	(C/Java-style) or by name, or both at once if that seems most entertaining.
	<p>
	To the code in the function's body, both the input and output parameters
	behave as local variables.  It's just that the inputs happen to come
	initialized with useful information, and the outputs go on to whoever's
	calling the function.
	<p>
	What exactly the "body of code" will looks like is yet to be determined.  It may
	well look like your wildest dreams, or your worst nightmares, or both.  It may
	be possible to make the "Code" class very abstract, and have different subclasses
	with different syntactic flavors (i.e., one more procedural, another more functional,
	another using some ultra-funky dataflow thingers...).  Or maybe it will be raw
	6502 machine code.  Who can tell?
	<p>
	<font size=+1><b>Structure:</b></font>
	<ul>
		<li> name
		<li> owner
		<li> inputs
		<li> outputs
		<li> body
	</ul>
	<p>
	<font size=+1><b>Signature:</b></font>
	<ul>
		<li> name
		<li> signatures of inputs & outputs
	</ul>
	<p>
	See note on synchronization and concurrent read safety in {@link Container}.
	
	@author Paul Cantrell
	@version [Development version]
*/

public class Function
	extends NamedElement
	{
	///** Just a dummy for now */
	private abstract class Code extends Element { };
	
	//------------------------------------------------------
	//  Constants
	//------------------------------------------------------
	
	/**	Part of a function.
	 *	@see org.eidola.kernel.event.StructureChanged */
	
	static public final ContainerPart
		INPUTS  = new ContainerPart("inputs"),
		OUTPUTS = new ContainerPart("outputs"),
		BODY    = new ContainerPart("body");
	
	//------------------------------------------------------
	//  Constructors
	//------------------------------------------------------
	
	/** Creates a new empty abstract function that takes nothing, does nothing with
	 *  it, and gives you nothing back, yet still will probably run a bit slowly.... */
	
	public Function()
		{
		inputs = new ArrayList();
		outputs = new ArrayList();
		parameterListener = new EventListener()
			{
			public void handleEvent(Event event)
				{ handleContentEvent((ContainerEvent) event); }
			};
		}
	
	//------------------------------------------------------
	//  Structural properties
	//------------------------------------------------------
	
	/** Returns the list of input parameters.
	 *  @return A list of {@link Variable}s. */
	
	public List/*<Variable>*/ getInputs()
		{ return inputs; }
		
	/** Returns the list of output parameters.
	 *  @return A list of {@link Variable}s. */
	
	public List/*<Variable>*/ getOutputs()
		{ return outputs; }
	
	/** Adds an input parameter at the end of the parameter list. */
	
	public void addInput(Variable input)
		throws IllegalElementType
		{ addParameter(input, true); }
	
	/** Adds an output parameter at the end of the parameter list. */
	
	public void addOutput(Variable output)
		throws IllegalElementType
		{ addParameter(output, false); }
	
	/** Adds a parameter at the end of the parameter list.
	 *  @param parameter The parameter to add
	 *  @param in True = input parameter; false = output parameter */
	
	public void addParameter(Variable parameter, boolean in)
		throws IllegalElementType
		{
		checkParameter(parameter);
		
		List newParameters = new ArrayList(in ? inputs : outputs);
		newParameters.add(parameter);
		newParameters = Collections.unmodifiableList(newParameters);
		if(in)
			inputs = newParameters;
		else
			outputs = newParameters;
		
		parameter.addListener(parameterListener, Engine.getInstance().getPropagatorQ());
		
		StructureChanged structEvent =
			new StructureChanged(this, in ? INPUTS : OUTPUTS);
		broadcastEvent(structEvent);
		broadcastEvent(new SignatureChanged(this, structEvent));
		}
	
	/** Removes an input parameter. */
	
	public void removeInput(Variable input)
		{ removeParameter(input, true); }
	
	/** Removes an output parameter. */
	
	public void removeOutput(Variable output)
		{ removeParameter(output, false); }
	
	/** Removes a parameter.
	 *  @param parameter The parameter to remove
	 *  @param in True = input parameter; false = output parameter */
	
	public void removeParameter(Variable parameter, boolean in)
		{
		if(parameter == null)
			throw new NullPointerException("parameter == null");
			
		List newParameters = new ArrayList(in ? inputs : outputs);
		newParameters.remove(parameter);
		newParameters = Collections.unmodifiableList(newParameters);
		if(in)
			inputs = newParameters;
		else
			outputs = newParameters;
		
		if(!inputs.contains(parameter) && !outputs.contains(parameter))
			parameter.removeListener(parameterListener);
		
		StructureChanged structEvent =
			new StructureChanged(this, in ? INPUTS : OUTPUTS);
		broadcastEvent(structEvent);
		broadcastEvent(new SignatureChanged(this, structEvent));
		}
	
	/** Sets the input parameters.
	 *  @param inputs A list of {@link Variable}s.  This method creates
	 *  	an internal copy, so it is safe to pass a list and subsequently modify it. */
	
	public void setInputs(List/*<Variable>*/ inputs)
		throws IllegalElementType
		{ setParameters(inputs, true); }
	
	/** Sets the output parameters.
	 *  @param inputs A list of {@link Variable}s.  This method creates
	 *  	an internal copy, so it is safe to pass a list and subsequently modify it. */
	
	public void setOutputs(List/*<Variable>*/ outputs)
		throws IllegalElementType
		{ setParameters(outputs, false); }
	
	/** Sets the input or output parameters.
	 *  @param inputs A list of {@link Variable}s.  This method creates
	 *  	an internal copy, so it is safe to pass a list and subsequently modify it.
	 *  @param in True = input parameters; false = output parameters */
	
	public void setParameters(List/*<Variable>*/ newParameters, boolean in)
		throws IllegalElementType
		{
		if(newParameters == null)
			newParameters = new LinkedList();
		
		for(Iterator i = newParameters.iterator(); i.hasNext(); )
			checkParameter(i.next());
		
		Set allParams = new HashSet(inputs);
		allParams.addAll(outputs);
		Set allNewParams = new HashSet(newParameters);
		allNewParams.addAll(in ? outputs : inputs);
		EventBroadcaster.updateBroadcasters(
			parameterListener,
			Engine.getInstance().getPropagatorQ(),
			allParams,
			allNewParams);
		
		newParameters = Collections.unmodifiableList(new ArrayList(newParameters));
		if(in)
			inputs = newParameters;
		else
			outputs = newParameters;
		
		StructureChanged structEvent =
			new StructureChanged(this, in ? INPUTS : OUTPUTS);
		broadcastEvent(structEvent);
		broadcastEvent(new SignatureChanged(this, structEvent));
		}
	
	/** Returns the body of code which describes the action this function takes,
	 *  or null if the function is purely abstract. */
	
	public Code getBody()
		{ return body; }
	
	//------------------------------------------------------
	//  Changes and events
	//------------------------------------------------------
	
	/** Handles an event passed on from content.  This handler
	 *  passes {@link SignatureChanged} events from parameters,
	 *  and calls {@link Element#handleContentEvent(ContainerEvent)}. */
	
	public void handleContentEvent(ContainerEvent event)
		{
		super.handleContentEvent(event);

		if(event instanceof CompileCompleted)
			{
			CompileRequired prop = (CompileRequired) event.getPropagatedFrom();
			if(prop instanceof SignatureChanged)
				synchronized(this)
					{ broadcastEvent(new SignatureChanged(this, prop)); }
			}
		}
	
	//------------------------------------------------------
	//  Semantic conditions and errors
	//------------------------------------------------------
	
	protected Container.Compilation makeNewCompilation()
		{ return new Compilation(); }
		
	public class Compilation
		extends NamedElement.Compilation
		{
		/** Returns true if <code>other</code> is a function with the same number of
		 *  input and output parameters, and each parameter of this function specializes
		 *  a corresponding parameter of <code>other</code> with the same name. */
		
		public boolean specializes(NamedElement other)
			{
			if(!(other instanceof Function))
			 	return false;
			Function otherFunc = (Function) other;
			return varListSpecializes(getInputs(),  otherFunc.getInputs())
				&& varListSpecializes(getOutputs(), otherFunc.getOutputs());
			}
		
		private boolean varListSpecializes(List varList1, List varList2)
			{
			Iterator i, i2;
			for(i = varList1.iterator(), i2 = varList2.iterator(); i.hasNext() && i2.hasNext(); )
				{
				Variable v1 = (Variable) i.next(), v2 = (Variable) i2.next();
				if(!((Variable.Compilation) v1.getCompilation()).specializes(v2))
					return false;
				}
			if(i.hasNext() || i2.hasNext())
				return false;
			return true; 
			}
		
		/** Calculates the contents of this function. */
		
		protected void calculateDerived()
			{
			contents = new HashSet();
			contents.addAll(inputs);
			contents.addAll(outputs);
			if(getBody() != null)
				contents.add(getBody());
			contents = Collections.unmodifiableSet(contents);
			
			elementsUsed = new HashSet(contents);
			elementsUsed.add(Function.this);
			elementsUsed.add(getOwner());
			elementsUsed = Collections.unmodifiableSet(elementsUsed);
			}
		
		/** Check the collection structure.  A Function must satisfy
		 *  the following conditions, in addition to the conditions required by
		 *  {@link Element.Compilation#checkLazy()}:
		 *  <ul>
		 *  	<li> This function must own its parameters and its body.
		 *  	<li> The names of all parameters must be unique.
		 *  	<li> No parameter can appear twice in either the input or output list
		 *  		(although it can appear once in each).
		 *  </ul>
		 *  @return A list of {@link SemanticViolation}s, or an empty list if the structure is OK. */
		
		protected void checkLazy()
			{
			super.checkLazy();
			
			// Check owners and element uniqueness
			
			checkParameters(inputs, INPUTS);
			checkParameters(outputs, OUTPUTS);
			
			// Check body's owner
			
			Code body = getBody();
			if(body != null && body.getOwner() != Function.this)
				addError(new IllegalContent(getBody(), body + " cannot be the body of " + Function.this));
			}

		private void checkParameters(List/*<Variable>*/ params, ContainerPart curPart)
			{
			Set paramSet = new HashSet();
			for(Iterator i = params.iterator(); i.hasNext(); )
				{
				Variable cur = (Variable) i.next();
				
				// Owner must match
				
				if(cur.getOwner() != Function.this)
					addError(new IllegalContent(cur, cur + " cannot be a parameter of " + Function.this));
			
				// Param cannot be an input or an output twice
				
				if(!paramSet.contains(cur))
					paramSet.add(cur);
				else
					addError(new DuplicateParameter(
						cur,
						curPart,
						cur + " cannot appear twice in the " + curPart.getName() + "!"));
				}
			}
		}		
	
	//------------------------------------------------------
	//  Debugging
	//------------------------------------------------------
	
	public void dump(int i)
		{
		Util.indent(i);
		
		System.out.print("Function " + toString());
		if(!getCompilation().isCurrent())
			System.out.print("  ** compile needed **");
		System.out.print('\n');
		
		Util.indent(i+2);
		System.out.print("inputs:\n");
		for(Iterator elem = getInputs().iterator(); elem.hasNext(); )
			((Element) elem.next()).dump(i+4);
		
		Util.indent(i+2);
		System.out.print("outputs:\n");
		for(Iterator elem = getOutputs().iterator(); elem.hasNext(); )
			((Element) elem.next()).dump(i+4);
		}

	//------------------------------------------------------
	//  Private
	//------------------------------------------------------
	
	private List inputs, outputs;
	private Code body;
	private EventListener parameterListener;
	
	private void checkParameter(Object param)
		throws IllegalElementType
		{
		if(param == null || !(param instanceof Variable))
			throw new IllegalElementType(
				(param instanceof Element) ? (Element) param : null,
				param + " cannot be a parameter, because it is not a variable.");
		}
	}
