/*______________________________________________________________________________
 * 
 * org.eidola.kernel.NamedElement
 * 
 * Part of the Eidola Kernel Reference Implementation
 * See http://eidola.org for oodles of relevant fun!
 * 
 *______________________________________________________________________________
 * 
 * Copyright 1998, 2000-2001 Paul Cantrell
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License version 2, as published by the
 * Free Software Foundation. See the file LICENSE.html for more information.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY, including the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc. / 59 Temple
 * Place, Suite 330 / Boston, MA 02111-1307 / USA.
 *_______________________________________________________________________________
 */

package org.eidola.kernel;

import org.eidola.kernel.event.*;
import org.eidola.kernel.error.UninitializedPart;

/**
	An element which occupies a place in the namespace.
	<p>
	A named element has a <i>name</i> -- its own unique name within its
	owner.  It also has a <i>full name</i>, which is just the chain of
	names of its indirect owners.  See {@link #getName()}, {@link Element#getFullName()},
	and {@link #toString()}.
	<p>
	Named elements can {@link NamedElement.Compilation#specializes(NamedElement other) specialize}
	other named elements.  Element <b>B</b> specializes element <b>A</b> if <b>B</b> can be 
	used in a context that expect something looking like <b>A</b>.  This
	specializing construction forms the basis of Eidola's type and inheritence systems --
	see {@link Class} for more on this.
	<p>
	A named element has a <i>signature</i>, which is
	the parts of this element's structure or its contents' structures
	which will affect the result of {@link NamedElement.Compilation#specializes(NamedElement other)}
	for this element, plus the name.
	<p>
	<font size=+1><b>Structure:</b></font>
	<ul>
		<li> name
		<li> owner
	</ul>
	<p>
	<font size=+1><b>Signature:</b></font>
	<ul>
		<li> name
	</ul>
	<p>
	See note on synchronization and concurrent read safety in {@link Container}.
	
	@see Class
	@see Package
	@see Function
	@see Variable
	
	@author Paul Cantrell
	@version [Development version]
*/

public abstract class NamedElement
	extends Element
	{
	static public final ContainerPart NAME = new ContainerPart("name");
	
	//------------------------------------------------------
	//  Constructors
	//------------------------------------------------------
	
	/** Creates a new named element with an empty name. */
	
	public NamedElement()
		{
		super();
		name = "";
		}
			
	//------------------------------------------------------
	//  Structural properties
	//------------------------------------------------------
	
	/** Returns this element's name within its container.  For example,
	 *  the name of the element "package.SomeClass.f.x" is "x". */
	
	public String getName()
		{ return name; }
		
	/** Sets this element's name.  This changes the element's structure
	 *  and its signature.
	 *  You must synchronize on the element when calling this method. */
	
	public void setName(String newName)
		{
		if(newName == null || newName.length() == 0)
			throw new NullPointerException("newName is empty");
		
		if(name.equals(newName))
			return;
			
		// We should remove illegal characters -- identifiers should only have
		// alphanumerics and underscores.
		// ...Or maybe we should allow all this junk??  Wait until there's an
		// interface to see if that's workable.
		
		name = newName;
		
		StructureChanged structEvent = new StructureChanged(this, NAME);
		broadcastEvent(structEvent);
		broadcastEvent(new SignatureChanged(this, structEvent));
		}
	
	//------------------------------------------------------
	//  Changes and events
	//------------------------------------------------------
	
	/** Returns a modification count for this element's signature.  The
	 *  signature is the parts of this element's structure or its
	 *  contents' structures which will affect
	 *  {@link NamedElement.Compilation#specializes(NamedElement other) specialization}
	 *  on this element, plus the name of the element. */
	
	public long getSignatureVersion()
		{ return signatureVersion; }
	
	/** Broadcasts an event about changes in this element, incrementing
	 *  appropriate modification counts.
	 *  @param event The event to broadcast. */
	
	public void broadcastEvent(Event event)
		{
		if(event instanceof SignatureChanged)
			signatureVersion++;
		super.broadcastEvent(event);
		}
	
	//------------------------------------------------------
	//  Semantic conditions and errors
	//------------------------------------------------------
	
	public abstract class Compilation
		extends Element.Compilation
		{
		/** Determines whether this element can be used in a context which expects
		 *  something looking like otherElement. */
		
		public abstract boolean specializes(NamedElement otherElement);
		
		protected void checkLazy()
			{
			super.checkLazy();
			
			if(name == null || name.length() == 0)
				addError(new UninitializedPart(NamedElement.this,
					NamedElement.this + " is nameless", NAME));
			}
		}

	//------------------------------------------------------
	//  Private
	//------------------------------------------------------
	
	private String name;
	private long signatureVersion;
	}
	
	
