/*______________________________________________________________________________
 * 
 * org.eidola.kernel.Namespace
 * 
 * Part of the Eidola Kernel Reference Implementation
 * See http://eidola.org for oodles of relevant fun!
 * 
 *______________________________________________________________________________
 * 
 * Copyright 2001 Paul Cantrell
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License version 2, as published by the
 * Free Software Foundation. See the file LICENSE.html for more information.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY, including the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc. / 59 Temple
 * Place, Suite 330 / Boston, MA 02111-1307 / USA.
 *_______________________________________________________________________________
 */

package org.eidola.kernel;

import org.eidola.kernel.event.*;
import org.eidola.kernel.event.EventListener;
import org.eidola.kernel.error.*;
import org.eidola.util.Util;

import java.util.*;

/**
	The top container in the container hierarchy.  A namespace generally corresponds
	to a physical unit of a program -- the code within a namespace is stored in the
	same place and the same way.  Full names of elements are unique within a namespace.
	A namespace has a single {@link #getRoot() root} element, typically a package.
	
	@author Paul Cantrell
	@version [Development version]
*/

public class Namespace
	extends Container
	{
	/**	Part of an container.
	 *	@see org.eidola.kernel.event.StructureChanged */
	
	static public final ContainerPart ROOT = new ContainerPart("root");

	/** Creates a new empty namespace.  The new namespace won't be valid
	 *  until somebody gives it a root element. */
	
	public Namespace()
		{
		rootListener = new EventListener()
			{
			public void handleEvent(Event event)
				{ handleContentEvent((ContainerEvent) event); }
			};
		}
	
	/** Returns this namespace. */
	
	public Namespace getNamespace()
		{ return this; }
	
	/** Returns the root element of the namespace. */
	
	public Element getRoot()
		{ return root; }
	
	/** Changes the root element of the namespace. */
	
	public void setRoot(Element root)
		throws IllegalElementType
		{
		if(root == null)
			throw new IllegalElementType(root, "A namespace cannot have a null root");
		if(this.root != root)
			{
			if(root != null)
				root.removeListener(rootListener);
			this.root = root;
			root.addListener(rootListener, Engine.getInstance().getPropagatorQ());
			broadcastEvent(new StructureChanged(this, ROOT));
			}
		}
	
	//------------------------------------------------------
	//  Semantic conditions and errors
	//------------------------------------------------------
	
	protected Container.Compilation makeNewCompilation()
		{ return new Compilation(); }
		
	public class Compilation
		extends Container.Compilation
		{
		/** Calculates the contents of this namespace, which are,
		 *  of course, just the root element and nothing else. */
		
		protected void calculateDerived()
			{
			contents = (root != null)
				? Collections.singleton(root)
				: Collections.EMPTY_SET;
			}
		
		/** Checks that this namespace has a root element and owns it. */
		
		protected void checkLazy()
			{
			super.checkLazy();
			
			if(root == null)
				addError(
					new UninitializedPart(Namespace.this, Namespace.this + " is rootless", ROOT));
			else if(root.getOwner() != Namespace.this)
				addError(
					new IllegalContent(
						root, root + " cannot be the root of " + Namespace.this
						+ " because its owner is " + root.getOwner()));
			}
		}
	
	//------------------------------------------------------
	//  Debugging
	//------------------------------------------------------
	
	public void dump(int i)
		{
		Util.indent(i);
		System.out.print(toString());
		if(!getCompilation().isCurrent())
			System.out.print("  ** compile needed **");
		System.out.println("");

		root.dump(i+4);
		}
	
	//------------------------------------------------------
	//  Private
	//------------------------------------------------------
	
	private Element root;
	private EventListener rootListener;
	}
	
	
