/*______________________________________________________________________________
 * 
 * org.eidola.kernel.Variable
 * 
 * Part of the Eidola Kernel Reference Implementation
 * See http://eidola.org for oodles of relevant fun!
 * 
 *______________________________________________________________________________
 * 
 * Copyright 1998, 2000-2001 Paul Cantrell
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License version 2, as published by the
 * Free Software Foundation. See the file LICENSE.html for more information.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY, including the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc. / 59 Temple
 * Place, Suite 330 / Boston, MA 02111-1307 / USA.
 *_______________________________________________________________________________
 */

package org.eidola.kernel;

import org.eidola.kernel.event.*;
import org.eidola.kernel.event.EventListener;
import org.eidola.kernel.error.IllegalElementType;
import org.eidola.kernel.error.UninitializedPart;
import org.eidola.util.Util;
import java.util.*;

/**
	A named and typed reference to an object.
	<p>
	<font size=+1><b>Structure:</b></font>
	<ul>
		<li> name
		<li> owner
		<li> type
	</ul>
	<p>
	<font size=+1><b>Signature:</b></font>
	<ul>
		<li> name
		<li> type
	</ul>
	<p>
	See note on synchronization and concurrent read safety in {@link Container}.
	
	@author Paul Cantrell
	@version [Development version]
*/

public class Variable extends NamedElement
	{
	/**	Part of an element.
	 *	@see org.eidola.kernel.event.StructureChanged */
	
	static public final ContainerPart TYPE = new ContainerPart("type");
	
	/** Creates a new variable with no type. */
	
	public Variable()
		{
		typeListener = new EventListener()
			{
			public void handleEvent(Event event)
				{ handleTypeEvent((ContainerEvent) event); }
			};
		}
	
	/** Returns this variable's type. */
	
	public Class getType()
		{ return type; }
	
	/** Sets this class's type. */
	
	public void setType(Class newType)
		throws IllegalElementType
		{
		if(newType == null)
			throw new IllegalArgumentException("newType == null");
		if(type == newType)
			return;
		
		if(type != null)
			type.removeListener(typeListener);
		newType.addListener(typeListener, Engine.getInstance().getPropagatorQ());
		
		type = newType;
		
		StructureChanged structEvent = new StructureChanged(this, TYPE);
		broadcastEvent(structEvent);
		broadcastEvent(new SignatureChanged(this, structEvent));
		}
	
	/** Handles events coming from the variable's type. */
	
	public void handleTypeEvent(ContainerEvent event)
		{
		if(event instanceof GeneralizationsChanged)
			synchronized(this)
				{ broadcastEvent(new SignatureChanged(this, event)); }
		}
	
	//------------------------------------------------------
	//  Semantic conditions and errors
	//------------------------------------------------------
	
	protected Container.Compilation makeNewCompilation()
		{ return new Compilation(); }
		
	public class Compilation
		extends NamedElement.Compilation
		{
		/** Returns true if the other element is a variable with
		 *  a compatible type (i.e. other's type is a generalization of this variable's type). */
		
		public boolean specializes(NamedElement other)
			{
			if(other == Variable.this)
				return true;
			return
				other instanceof Variable
				&& getType() != null
				&& ((Variable) other).getType() != null
				&& ((Class.Compilation) getType().getCompilation()).specializes(((Variable) other).getType());
			}
		
		/** Calculates the contents of this function. */
		
		protected void calculateDerived()
			{
			contents = Collections.unmodifiableSet(new HashSet());
			
			elementsUsed = new HashSet();
			elementsUsed.add(Variable.this);
			elementsUsed.add(getOwner());
			elementsUsed.add(type);
			elementsUsed = Collections.unmodifiableSet(elementsUsed);
			}
		
		protected void checkLazy()
			{
			super.checkLazy();
			
			if(type == null)  // This check can go away if we have Object as the default type
				addError(new UninitializedPart(Variable.this,
					Variable.this + " needs a type.  Poor thing.", TYPE));
			}
		}
	
	public void dump(int i)
		{
		Util.indent(i);
				
		System.out.print("Variable "
						+ toString()
						+ " : "
						+ getType());
		if(!getCompilation().isCurrent())
			System.out.print("  ** compile needed **");
		System.out.print('\n');
		}
	
	private Class type;
	private EventListener typeListener;
	}
