/*______________________________________________________________________________
 * 
 * org.eidola.kernel.event.EventQueue
 * 
 * Part of the Eidola Kernel Reference Implementation
 * See http://eidola.org for oodles of relevant fun!
 * 
 *______________________________________________________________________________
 * 
 * Copyright 2000-2001 Paul Cantrell
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License version 2, as published by the
 * Free Software Foundation. See the file LICENSE.html for more information.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY, including the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc. / 59 Temple
 * Place, Suite 330 / Boston, MA 02111-1307 / USA.
 *_______________________________________________________________________________
 */

package org.eidola.kernel.event;

import java.util.*;

/**
	A queue dispatches {@link Event}s to {@link EventListener}s.  Typically,
	a single queue is associated with a listener or a collection of listeners
	with a single purpose -- i.e. compiling elements or updating a view.
	<p>
	There are two common ways of running an event queue:
	<ul>
		<li> As a single thread:
			<blockquote><code>
				EventQueue q = new EventQueue("test");<br>
				q.start();
			</code></blockquote>
		<li> Using multiple threads using {@link org.eidola.util.ThreadFlock}:
			<blockquote><code>
				EventQueue q = new EventQueue("test");<br>
				ThreadFlock flock = new ThreadFlock(q, 4);<br>
				flock.start();
			</code></blockquote>
	</ul>
	
	@see Event
	@see EventBroadcaster
	@see EventListener
	
	@author Paul Cantrell
	@version [Development version]
*/

public class EventQueue
	extends Thread
	implements Runnable
	{
	/** For debugging. */
	public String name;
	/** For debugging -- returns this queue's name. */
	public String toString() { return name; }
	
	/** Constructs a new empty queue.
	 *  @param name For debugging. */
	public EventQueue(String name)
		{
		this.name = name;
		setDaemon(true);
		eventq = new LinkedList();
//		eventSet = new HashSet(); // For fast redundancy checks
		}
	
	/** Determines whether this queue is empty. */
	
	public boolean isEmpty()
		{ return eventq.isEmpty(); }
	
	/** Adds an event to the queue.  Typically, you will not call this directly;
	 *  that's really the job of {@link EventBroadcaster#broadcastEvent(Event)}. */
	
	public void addEvent(Event event, EventListener target)
		{
		Entry newEntry = new Entry(event, target);
//		if(!eventSet.contains(newEntry)) // Don't post redundant events
			synchronized(eventq)
				{
				if(eventq.isEmpty())
					eventq.notify();
				eventq.addLast(newEntry);
//				eventSet.add(newEntry);
				if(debug)
					System.out.println("  #(q) Enqueued " + newEntry);
				}
//		else if(debug)
//			System.out.println("  #(q) Skipped redundant " + newEntry);
		}
	
	/** Consumes and handles events from this queue.  Typically, you will not call this
	 *  method directly, but rather call {@link Thread#start()}. */
	
	public void run()
		{
		try
			{
			while(!Thread.interrupted())
				{
				// Get an event
				Entry qe;
				synchronized(eventq)
					{
					while(eventq.isEmpty())
						eventq.wait();
					qe = (Entry) eventq.removeFirst();
//					eventSet.remove(qe);
					}
				
				// Handle it
				qe.target.handleEvent(qe.event);
				if(debug)
					System.out.println("  #(q) Handled " + qe);
				}
			}
		catch(InterruptedException ie)
			{ }
		catch(Throwable err)
			{ err.printStackTrace(System.err); }
		}
	
	private LinkedList eventq;
//	private Set eventSet;

	private class Entry
		{
		public Event event;
		public EventListener target;
		
		public Entry(Event event, EventListener target)
			{
			this.event = event;
			this.target = target;
			}
		
		public String toString()
			{ return "(" + event + ") for (" + target + ")"; }
		
		public int hashCode()
			{ return event.hashCode() * 61 + target.hashCode(); }
		
		public boolean equals(Object other)
			{
			if(other == null)
				return false;
			return (other instanceof Entry)
				&& ((Entry) other).target == target
				&& ((Entry) other).event.equals(event);
			}
		}

	static public boolean debug = false;
	}

