/*______________________________________________________________________________
 * 
 * org.eidola.util.ThreadFlock
 * 
 * Part of the Eidola Kernel Reference Implementation
 * See http://eidola.org for oodles of relevant fun!
 * 
 *______________________________________________________________________________
 * 
 * Copyright 2000-2001 Paul Cantrell
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License version 2, as published by the
 * Free Software Foundation. See the file LICENSE.html for more information.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY, including the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc. / 59 Temple
 * Place, Suite 330 / Boston, MA 02111-1307 / USA.
 *_______________________________________________________________________________
 */

package org.eidola.util;

import java.util.*;

/**
	Creates a collection of threads running the same {@link Runnable}.
	The threads have the same priority and the same daemon status.
	You can start, interrupt, and join the threads en masse.
	
	@author Paul Cantrell
	@version [Development version]
*/

public class ThreadFlock
	{
	/** Creates a new flock with one thread.
	 *  To start the threads, use {@link #start()}. */
	
	public ThreadFlock(Runnable runnable)
		{
		this(runnable, 1);
		}
	
	/** Creates a new flock with an arbitrary number of threads.
	 *  To start the threads, use {@link #start()}. */
	
	public ThreadFlock(Runnable runnable, int threadCount)
		{
		this.runnable = runnable;
		threads = new ArrayList();
		setPriority(Thread.NORM_PRIORITY);
		setThreadCount(threadCount, false);
		}
	
	/** Creates a new flock with one thread.
	 *  The flock takes the priority and daemon status of the thread parameter,
	 *  but that thread does not actually become a member of the flock.
	 *  To start the threads, use {@link #start()}. */
	
	public ThreadFlock(Thread thread)
		{
		this(thread, 1);
		}
	
	/** Creates a new flock with an arbitrary number of threads.
	 *  The flock takes the priority and daemon status of the thread parameter,
	 *  but that thread does not actually become a member of the flock.
	 *  To start the threads, use {@link #start()}. */
	
	public ThreadFlock(Thread thread, int threadCount)
		{
		this((Runnable) thread, threadCount);
		setDaemon(thread.isDaemon());
		setPriority(thread.getPriority());
		}
	
	/** Returns the number of threads in this flock. */
	
	public int getThreadCount()
		{
		return threads.size();
		}
	
	/** Sets the number of threads in this flock.
	 *  If this is an increase, this method will create the new
	 *  threads, and will start them if the start parameter is true.  If this is a
	 *  decrease, this method will {@link java.lang.Thread#interrupt() interrupt}
	 *  the appropriate number of threads. */
	
	public void setThreadCount(int threadCount, boolean start)
		{
		if(threadCount < 0)
			throw new IllegalArgumentException("threadCount < 0");
		
		int oldCount = threads.size();
		if(threadCount == oldCount)
			return;
		
		for(int n = threadCount; n < oldCount; n++)
			{
			((Thread) threads.get(n)).interrupt();
			threads.remove(n);
			}
		
		for(int n = oldCount; n < threadCount; n++)
			{
			Thread t = new Thread(runnable);
			threads.add(t);
			if(start)
				t.start();
			}
		
		setPriority(priority);
		}
	
	/** Returns the priority of the threads in this flock.  Note that each
	 *  individual thread has this priority, so a flock with 50 threads of normal
	 *  priority is actually an extremely high-priority flock.
	 *  @see java.lang.Thread#getPriority() */
	
	public int getPriority()
		{
		return priority;
		}
	
	/** Sets the priority of all the threads in this flock.  This affects
	 *  threads whether or not they are running, and affects any threads
	 *  created by subsequent calls to {@link #setThreadCount(int,boolean) setThreadCount}.
	 *  @see java.lang.Thread#setPriority(int) */
	
	public void setPriority(int priority)
		{
		this.priority = priority;
		for(int n = 0; n < threads.size(); n++)
			((Thread) threads.get(n)).setPriority(priority);
		}
	
	/** Determines whether the threads in this flock are daemons.
	 *  (Daemon threads die automatically when the Java VM goes away.)
	 *  @see java.lang.Thread#isDaemon() */
	
	public boolean isDaemon()
		{
		return daemon;
		}
	
	/** Changes whether the threads in this flock are daemons.  This affects
	 *  threads whether or not they are running, and affects any threads
	 *  created by subsequent calls to {@link #setThreadCount(int,boolean) setThreadCount}.
	 *  @see java.lang.Thread#setDaemon(boolean) */
	
	public void setDaemon(boolean daemon)
		{
		this.daemon = daemon;
		for(int n = 0; n < threads.size(); n++)
			((Thread) threads.get(n)).setDaemon(daemon);
		}
	
	/** Starts all the threads in the flock.
	 *  @see java.lang.Thread#start() */
	
	public void start()
		{
		for(int n = 0; n < threads.size(); n++)
			((Thread) threads.get(n)).start();
		}
	
	/** Interrupts all the threads in the flock.
	 *  @see java.lang.Thread#interrupt() */
	
	public void interrupt()
		{
		for(int n = 0; n < threads.size(); n++)
			((Thread) threads.get(n)).interrupt();
		}
	
	/** Waits for all the threads in the flock to die.
	 *  @see java.lang.Thread#join(long) */
	
	public void join(long time)
		throws InterruptedException
		{
		for(int n = 0; n < threads.size(); n++)
			((Thread) threads.get(n)).join(time);
		}
	
	private Runnable runnable;
	private ArrayList threads;
	private int priority;
	private boolean daemon;
	}


